package QEMU::Manager;

use base qw(Net::DBus::Object);

use Net::DBus::Exporter qw(com.redhat.qemu.Manager);

use QEMU::Manager::Machine;
use QEMU::Manager::ISOImage;
use QEMU::Manager::DiskImage;
use QEMU::Manager::Formatter;

use File::Spec::Functions qw(catfile catdir);
use File::Path;
use IO::File;

our $VERSION = "0.0.5";

sub new {
    my $class = shift;
    my $service = shift;
    my $self = $class->SUPER::new($service, "/qemu/manager");
    my $configDir = shift;

    $self->{machines} = {};
    $self->{isos} = {};
    $self->{disks} = {};
    $self->{configDir} = $configDir;

    bless $self, $class;
    
    $self->_load_machines();
    $self->_load_isoimages();
    $self->_load_diskimages();

    $self->{formatter} = QEMU::Manager::Formatter->new($self);

    return $self;
}


sub _load_machines {
    my $self = shift;
    
    my $dir = $self->get_config("machines");
    opendir DIR, $dir
	or die "cannot read machines directory: $!";
    
    my @machines = map { s/\.conf$//; $_ } grep { /^[_a-z0-9]+\.conf$/i } readdir DIR;

    closedir DIR
	or die "cannot close directory: $!";
    
    foreach my $machine (@machines) {
	$self->_add_machine($machine);
    }
}


sub _load_isoimages {
    my $self = shift;
    
    my $dir = $self->get_config("isoimages");
    opendir DIR, $dir
	or die "cannot read iso images directory: $!";
    
    my @isoimages = map { s/\.conf$//; $_ } grep { /^[_a-z0-9]+\.conf$/i } readdir DIR;

    closedir DIR
	or die "cannot close directory: $!";
    
    foreach my $isoimage (@isoimages) {
	$self->_add_iso_image($isoimage);
    }
}


sub _load_diskimages {
    my $self = shift;
    
    my $dir = $self->get_config("diskimages");
    opendir DIR, $dir
	or die "cannot read disk images directory: $!";
    
    my @diskimages = map { s/\.conf$//; $_ } grep { /^[_a-z0-9]+\.conf$/i } readdir DIR;

    closedir DIR
	or die "cannot close directory: $!";
    
    foreach my $diskimage (@diskimages) {
	$self->_add_disk_image($diskimage);
    }
}


sub get_config {
    my $self = shift;
    my $type = shift;

    my $base = catdir($self->{configDir}, $type);
    mkpath([$base]);

    if (@_) {
	my $name = shift;
	my $ext = shift;
	return catfile($base, $name . $ext);
    } else {
	return $base;
    }
}

dbus_method("list_machines", [], [["array", "string"]]);
sub list_machines {
    my $self = shift;
    
    my @machines = keys %{$self->{machines}};
    return \@machines;
}

dbus_method("has_machine", ["string"], ["bool"]);
sub has_machine {
    my $self = shift;
    my $name = shift;
    
    return exists $self->{machines}->{$name} ? 1 : 0;
}

dbus_signal("machine_created", ["string"]);
dbus_method("create_machine", ["string"]);
sub create_machine {
    my $self = shift;
    my $name = shift;
    
    die "name must only consist of characters a-z, 0-9, _" 
	unless $name =~ /^[_a-z0-9]+$/i;
    
    die "a machine with name '$name' already exists"
	if $self->{machines}->{$name};

    my $config = $self->get_config("machines", $name, ".conf");
    my $fh = IO::File->new(">$config")
	or die "cannot create machine config $config: $!";
    
    $fh->close
	or die "cannot save machine config $config: $!";
    
    my $machine = $self->_add_machine($name);
    $self->emit_signal("machine_created", $name);
}

dbus_signal("machine_deleted", ["string"]);
dbus_method("delete_machine", ["string"]);
sub delete_machine {
    my $self = shift;
    my $name = shift;

    die "no machine with name '$name' currently exists"
	unless $self->{machines}->{$name};
    
    $self->_remove_machine($name);

    my $config = $self->get_config("machines", $name, ".conf");
    unlink $config
	or die "cannot remove machine config $config: $!";
    $self->emit_signal("machine_deleted", $name);
}

sub _add_machine {
    my $self = shift;
    my $name = shift;
    
    my $machine = QEMU::Manager::Machine->new($self, 
					      $name, 
					      $self->get_config("machines", $name, ".conf"),
					      $self->get_config("logs", $name, ".log"));
    
    $self->{machines}->{$name} = $machine;
}

sub _remove_machine {
    my $self = shift;
    my $name = shift;
    
    my $machine = $self->{machines}->{$name};
    $machine->disconnect;
    delete $self->{machines}->{$name};
}


dbus_method("list_iso_images", [], [["array", "string"]]);
sub list_iso_images {
    my $self = shift;
    
    my @images = keys %{$self->{isoimages}};
    return \@images;
}


dbus_method("add_iso_image", ["string", "string", "string"]);
dbus_signal("iso_image_added", ["string"]);
sub add_iso_image {
    my $self = shift;
    my $name = shift;
    my $title = shift;
    my $path = shift;

    die "name must only consist of characters a-z, 0-9, _" 
	unless $name =~ /^[_a-z0-9]+$/i;
    
    if (exists $self->{isoimages}->{$name}) {
	die "an iso image with name $name already exists";
    }
    
    my $image = $self->_add_iso_image($name);
    $image->set_filename($path);
    $image->set_title($title);
    $image->save_config();
    
    $self->emit_signal("iso_image_added", $name);
}

dbus_method("remove_iso_image", ["string"]);
dbus_signal("iso_image_removed", ["string"]);
sub remove_iso_image {
    my $self = shift;
    my $name = shift;
    
    unless (exists $self->{isoimages}->{$name}) {
	die "an iso image with name $name does not exist";
    }
    
    unlink $self->get_config("isoimages", $name, ".conf")
	or die "cannot remove config file for iso image $name: $!";
    
    delete $self->{isoimages}->{$name};
    
    $self->emit_signal("iso_image_removed", $name);
}

sub _add_iso_image {
    my $self = shift;
    my $name = shift;
    
    my $image = QEMU::Manager::ISOImage->new($self,
					     $name,
					     $self->get_config("isoimages", $name, ".conf"));
    
    $self->{isoimages}->{$name} = $image;
    
    return $image;
}					     

dbus_method("list_disk_images", [], [["array", "string"]]);
sub list_disk_images {
    my $self = shift;

    my @images = keys %{$self->{diskimages}};
    return \@images;
}

dbus_method("add_disk_image", ["string", "string", "string"]);
dbus_signal("disk_image_added", ["string"]);
sub add_disk_image {
    my $self = shift;
    my $name = shift;
    my $title = shift;
    my $path = shift;

    die "name must only consist of characters a-z, 0-9, _" 
	unless $name =~ /^[_a-z0-9]+$/i;
    
    if (exists $self->{diskimages}->{$name}) {
	die "an disk image with name $name already exists";
    }
    
    my $image = $self->_add_disk_image($name);
    $image->set_filename($path);
    $image->set_title($title);
    $image->save_config();
    
    $self->emit_signal("disk_image_added", $name);
}

dbus_method("remove_disk_image", ["string"]);
dbus_signal("disk_image_removed", ["string"]);
sub remove_disk_image {
    my $self = shift;
    my $name = shift;
    
    unless (exists $self->{diskimages}->{$name}) {
	die "an disk image with name $name does not exist";
    }
    
    unlink $self->get_config("diskimages", $name, ".conf")
	or die "cannot remove config file for disk image $name: $!";
    
    delete $self->{diskimages}->{$name};
    
    $self->emit_signal("disk_image_removed", $name);
}

sub _add_disk_image {
    my $self = shift;
    my $name = shift;
    
    my $image = QEMU::Manager::DiskImage->new($self,
					      $name,
					      $self->get_config("diskimages", $name, ".conf"));
    
    $self->{diskimages}->{$name} = $image;
    
    return $image;
}					     


1;
