package QEMU::Manager::Formatter;

use strict;
use warnings;

use base qw(Net::DBus::Object);
use Net::DBus::Exporter qw(com.redhat.qemu.manager.Formatter);

use Config::Record;
use IPC::Open3;
use POSIX qw(F_GETFL F_SETFL  WNOHANG :errno_h strftime);
use Fcntl;


sub new {
    my $class = shift;
    my $machine = shift;
    my $name = shift;
    my $self = $class->SUPER::new($machine,
				  "/formatter");

    bless $self, $class;

    return $self;
}


dbus_method("create_image", ["string", "string", "uint32"]);
sub create_image {
    my $self = shift;
    my $output = shift;
    my $format = shift;
    my $size = shift;
    
    if (-e $output) {
	die "output file '$output' already exists";
    }

    if ($self->is_running) {
	die "operation is already in progress";
    }

    my @cmd = ("qemu-img", "create", "-f", $format, $output, $size . "M");
    
    $self->_start_operation(@cmd);
    $self->emit_signal("operation_started", "convert");

    $self->{timer} = Net::DBus::Reactor->main->add_timeout(250,
							   Net::DBus::Callback->new(object => $self,
										    method => "_operation_stepped",
										    args => [$output, $size]));

}




dbus_method("convert_image", ["string","string", "string"]);
sub convert_image {
    my $self = shift;
    my $input = shift;
    my $output = shift;
    my $format = shift;
    
    if (-e $output) {
	die "output file '$output' already exists";
    }

    if (!-e $input) {
	die "input file '$input' does not exist";
    }

    if ($self->is_running) {
	die "operation is already in progress";
    }


    my @cmd = ("qemu-img", "convert", $input, "-O", $format, $output);
    
    $self->_start_operation(@cmd);
    $self->emit_signal("operation_started", "convert");

    my $size = (stat $input)[7];

    $self->{timer} = Net::DBus::Reactor->main->add_timeout(250,
							   Net::DBus::Callback->new(object => $self,
										    method => "_operation_stepped",
										    args => [$output, $size]));

}


dbus_signal("operation_started", ["string"]);
sub _start_operation {
    my $self = shift;
    my @cmd = @_;

    my ($stdin, $stdout, $stderr);
    $stdin = IO::Handle->new();
    $stdout = IO::Handle->new();
    $stderr = IO::Handle->new();
    
    my $pid = open3($stdin, $stdout, $stderr, @cmd);
    # Nothing of importance on stdin
    $stdin->close;

    $self->{pid} = $pid;
    $self->{stdout} = $stdout;
    $self->{stderr} = $stderr;
    $self->_setup_handle_nonblock($stdout);
    $self->_setup_handle_nonblock($stderr);

    my $reactor = Net::DBus::Reactor->main;
    $reactor->add_read($stdout->fileno, Net::DBus::Callback->new(object => $self,
								 method => "_handle_data",
								 args => [ $stdout ]));
    $reactor->add_read($stderr->fileno, Net::DBus::Callback->new(object => $self,
								 method => "_handle_data",
								 args => [ $stderr ]));

}

dbus_signal("operation_stepped", ["uint32"]);
sub _operation_stepped {
    my $self = shift;
    my $file = shift;
    my $size = shift;

    return unless $self->{pid};
    
    # XXX this doesn't actually work, because it appears qemu-img
    # seeks right to end of file immediately, making it appear
    # 100% in size even though its not copied the data yet
    if (-e $file) {
	my $complete = (stat _)[7];
	my $percentage = $complete * 100 / $size;
	if ($percentage > 100) {
	    $percentage = 100;
	}

	$self->emit_signal("operation_stepped", $percentage);
    } else {
	$self->emit_signal("operation_stepped", 0);
    }
}

dbus_method("is_running", [], ["bool"]);
sub is_running {
    my $self = shift;
    return $self->{pid} ? 1 : 0;
}

dbus_signal("operation_aborted");
dbus_method("abort_operation");
sub abort_operation {
    my $self = shift;
    
    $self->_kill;
    $self->emit_signal("operation_aborted");
}

sub _kill {
    my $self = shift;

    kill &POSIX::SIGTERM, $self->{pid};
    my $got = waitpid $self->{pid}, WNOHANG;
    if ($got != $self->{pid}) {
	kill &POSIX::SIGKILL, $self->{pid};
	sleep 1;
	$got = waitpid $self->{pid}, WNOHANG;
    }

    if ($self->{stdout}) {
	my $reactor = Net::DBus::Reactor->main;
        $reactor->remove_read($self->{stdout}->fileno);
	# Suck down all pending data
	while ($self->_discard_data($self->{stdout}, $self->{log}) == 1) {};
	$self->{stdout}->close;
	delete $self->{stdout};
    }

    if ($self->{stderr}) {
	my $reactor = Net::DBus::Reactor->main;
        $reactor->remove_read($self->{stderr}->fileno);
	# Suck down all pending data
	while ($self->_discard_data($self->{stderr}, $self->{log}) == 1) {};
	$self->{stderr}->close;
	delete $self->{stderr};
    }
    
    delete $self->{pid};

    if ($self->{timer}) {
	Net::DBus::Reactor->main->remove_timeout($self->{timer});
	delete $self->{timer};
    }
}

sub _setup_handle_nonblock {
    my $self = shift;
    my $handle = shift;
    
    my $flags = 0;
    fcntl($handle, F_GETFL, $flags)
        or die "Couldn't get flags for handle : $!";
    
    $flags |= O_NONBLOCK;
    fcntl($handle, F_SETFL, $flags)
        or die "Couldn't set flags for handle: $!";
}


dbus_signal("operation_completed");
sub _handle_data {
    my $self = shift;
    my $handle = shift;

    # During cleanup there is a chance a calback will be invoked
    # by the reactor even after we're all done, so this is an
    # important sanity check
    return unless $self->is_running;

    my $ret = $self->_discard_data($handle);
    if ($ret == 0) {
        $self->_kill($handle);
	$self->emit_signal("operation_completed");
    }
}


sub _discard_data {
    my $self = shift;
    my $src = shift;

    my $buffer = '';
    my $rv = $src->sysread($buffer, 4096);
    if (!defined($rv)) {
        if ($! == EAGAIN) {
            # would block
            return -1;
        } else {
            die "cannot read from pipe " . $src->fileno . ": $!";
        }
    } else {
        if ($rv) {
            # successfully read $rv bytes from HANDLE
	    # now just discard it
	    # XXX log it via log4perl ?
            return 1;
        } else {
            # End of file
            return 0;
        }
    }
}

